/*
 * Author:    Konrad Rzeszutek <conradus@iname.com>
 * Date:      Sept 21th, 1999 - Sept 24th 1999.
 * File:      Editor2D/DrawCircle.java
 *
 * Notes:
 *   This class serves to draw a circle using the Midpoint Formula.
 *   The circle can either by thick, patterned, both, or none of these.
 *   From the AWT, it only uses drawRect to draw a single point.
 *   This class also has a switch (fill) which will fill the circle.
 *
 */
package Editor2D;

import java.awt.*;


public class DrawCircle extends MyPrimitives {

    /*
     * Initializes the DrawCircle object
     *
     * require:
     *   OffG != null  && canvas != null &&
     *   ( (state == MyPrimitives.DRAW_NO_PATTERN) || (state == MyPrimitives.DRAW_PATTERN_ON)
     *   || (state == MyPrimitives.DRAW_PATTERN_OFF) || (state == MyPrimitives.DRAW_THICK_ON)
     *   || (state == MyPrimitives.DRAW_THICK_OFF))
     */
    public DrawCircle(Graphics OffG, Canvas canvas, int state) {

	super();
	this.canvas = canvas;		
	this.OffG = OffG;

	maxPoints = 2;   // The max points we will use
	myState = state;
	xPoints = new int[maxPoints]; // Our X-points "holders"
	yPoints = new int[maxPoints]; // Our Y-points "holders"
	
    }
    
    /*
     * Draws the circle on OffG.
     * 
     * Note:
     *  if ((canAdd()) || (isDeleted())) then draw() will not draw anything.
     */     
    public void draw () {
	
	int[] XY = new int[maxPoints]; // Our X,Y points.
	
	int gM = 0; // fM = gM - 1/4.
	
	if ((nPoints == maxPoints) && (!deleted)) {  // Do we have enough points and are we exterminated?

   	    R = radius();
	    XY[0]= 0; // We will use our Cartesian coordiante system. 
	    XY[1]= R; // The central point y+R will server as a starting point. drawPoint fixes the points back to the Java coordinate system.
	    patternC = 0; // Used in case we want to draw a dotted circle
	    gM = 1 - R;
	    drawPoints(XY); 
	    while (Math.abs(XY[0]) < Math.abs(XY[1])) {		   
		if (gM < 0) { // Shall we go East?
		    gM += 2*XY[0]+3; 
		} else { // No, we will go NE.
		    gM += 2*XY[0]-2*XY[1] + 5;
		    XY[1] --; // We decrease Y instead of increasing it b/c of Java's coordiante system
		    
		}
		XY[0]++;
		drawPoints(XY); // Draw 8 points (in 8 diff quadrants).
		patternC++;     // Used in case user wants the circle to be in zebra patterns.
	    }

	    
	}
    }

    /*
      * Calculate the distance of the centrum of the circle to the point pt.
      * (Formula: d = (x2-pt.x)^2 + (y2-pt.y)^2) -- note, the d isn't
      * squared. This is done to speed up the process (all other classes don't
      * square their distance either).
      *
      * require:
      *   pt != null
      *
      * return:
      *   if (isDeleted) then MyPrimitives.POS_INFINITY, else correct distance (without squaring)
      */
    public int distance(Point pt) {
	
	int[] v = new int[maxPoints];
	if (!deleted) { 
	    v[0] = xPoints[0] - pt.x; // (x2-pt.x)
	    v[0] *= v[0];             // (x2-pt.x)^2
	    v[1] = yPoints[0] - pt.y; // The same thing for y as above
	    if (v[1] ==0) v[1]=1;     // Just in case we get zero -- we don't want div by zero.
	    v[1] *= v[1];
	    
	    return ((v[0]+v[1])-radius());       // return unsquared distance
	}
	return MyPrimitives.POS_INFINITY;             // else just return a pos infinity.
    }
    
    /*
      * "Dumps" information for the user
      * 
      * return:
      *   dump() != null
      */
    public String dump () {
    
	return "\nOperation: circle (radius:"+R+")"+super.toString();
    }


    /*
     * Tells whether this is a circle or not.
     * required for DrawFill.java
     *
     * return:
     *  true
     */
    public boolean isCircle() {

	return true;

    }
    
    /*
     * Fills the circle based on the state.
     *
     * require:
     *     ((state == MyPrimitives.DRAW_NO_PATTERN) || (state == MyPrimitives.DRAW_PATTERN_ON)
     *   || (state == MyPrimitives.DRAW_PATTERN_OFF) || (state == MyPrimitives.DRAW_THICK_ON)
     *   || (state == MyPrimitives.DRAW_THICK_OFF))
     *
     */
    public void fill (int state) {
	
	fill_state = state; 
	fillCircle = true; // This will cause draw() to fill instead of drawing the outline.
	this.draw();
    }
    
    // Couple of private methods used for drawing circles.
    
    /*
     * Draws a atomic/single point.
     *
     * note:
     *  This method will convert x and y into Java's coordinate system.
     *
     * require:
     *  x && y be in a Cartesian coordinate system.
     *
     */
    private void drawPoint (int x, int y) {
	
	int[] XYTmp = new int[maxPoints];      // Our Point holder
	int i,j;                               // Two counters.
	Color foreground = OffG.getColor();    // Our current color.
	Color background = canvas.getBackground(); // Our background color
	
	XYTmp[0] = (x + xPoints[0]);    // Convert to Java coordinate system
	XYTmp[1] = (-y + yPoints[0]); 
	
	if (fillCircle) {		// We need only to fill the circle
	    
	    // Right side of the circle.
	    for (i=xPoints[0];i<XYTmp[0];i++) {
		if ((fill_state & DRAW_PATTERN_ON) == DRAW_PATTERN_ON) 
		    // The colors are reveresed, otherwise nothing would be seen
		    if (((i+XYTmp[1]) / 3) * 3 == (i+XYTmp[1])) OffG.setColor(background);
				else OffG.setColor(foreground);				
		OffG.drawRect(i,XYTmp[1],1,0);
	    } 
	    // Left side of the circle.
	    for (i=xPoints[0];i>XYTmp[0];i--) {
		if ((fill_state & DRAW_PATTERN_ON) == DRAW_PATTERN_ON) 
		    if (((i+XYTmp[1]-1) / 3) * 3 == (i+XYTmp[1]-2)) OffG.setColor(background);
		    else OffG.setColor(foreground); 
		OffG.drawRect(i,XYTmp[1],1,0);		
	    }
	} 
	else {
	    // Checks to see if we have to draw patterned or thick points.
	    if ((myState & DRAW_PATTERN_ON) == DRAW_PATTERN_ON)
		if ((patternC % 4) > 2) 
		    OffG.setColor(background); // Sets the color of the background
	    
	    if ((myState & DRAW_THICK_ON) == DRAW_THICK_ON) 
		for (i=0;i<=3;i++)
		    for (j=0;j<=3;j++) // We are using a simple thicking way.
			OffG.drawRect(XYTmp[0]-1+i,XYTmp[1]-1+j,1,0);
	    
	    OffG.drawRect(XYTmp[0],XYTmp[1],1,0);
	    OffG.setColor(foreground); // Change the color back to the original color
	}
	OffG.setColor(foreground);
    }
    

    /*
      * Draws 8 points on OffG.
      *
      * require:
      *  xy.length() == 2 & xy values are in the Cartesian coordinate system.
      *
      */
    private void drawPoints (int[] XY) {
    
	// I quadrant (0-45)
	drawPoint(XY[1],XY[0]);  // swap.
	// II quadrant (45-90)
	drawPoint(XY[0],XY[1]);  // no changes.
	// III quadrant	(90-120)
	drawPoint(-XY[0],XY[1]); // negate Y.
	// IV quadrant (120-180)
	drawPoint(XY[1],-XY[0]); // swap and negate X.
	// V quadrant (180-225)
	drawPoint(-XY[1],-XY[0]);// swap and negate X and Y.
	// VI quadrant (225-270)
	drawPoint(-XY[0],-XY[1]);// negate X and Y.
	// VII quadrant (270-315)
	drawPoint(XY[0],-XY[1]); // negate Y.
	// VIII quadrant (315-360)
	drawPoint(-XY[1],XY[0]); // swap and negate Y.	
    }
   
    /*
     * Compute the radius of the circle.
     * 
     * return:
     *  0..n
     */ 
     private int radius() {
	
	 int[] v = new int[maxPoints];
	 if (R < 0) {
	     v[0] = xPoints[0] - xPoints[1]; // (x2-pt.x)
	    v[0] *= v[0];             // (x2-pt.x)^2
	    v[1] = yPoints[0] - yPoints[1]; // The same thing for y as above
	    if (v[1] ==0) v[1]=1;     // Just in case we get zero -- we don't want div by zero.
	    v[1] *= v[1];
	    
	    return ( (int)(Math.sqrt(v[0]+v[1])));       // return squared distance
	 } 
	 return R;  	    
     }


    private int patternC = 0;  // For checking the pattern
    private int R = -1;        // Radius
    private Graphics OffG;      // Where we will draw
    private Canvas canvas;     // From where we will get the background color.
    private boolean fillCircle = false; // We won't fill be default.
    private int fill_state;     // Filling pattern -- don't care at this state.
	      
}
